# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2023 Bryant E. McDonnell (See AUTHORS)
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""
Python extensions for the SWMM5 Programmers toolkit.
"""

# Standard library imports
import packaging.version
import sys
from datetime import datetime

# Third party imports
from swmm.toolkit import solver

# Local imports
import pyswmm.toolkitapi as tka


class SWMMException(Exception):
    """Custom exception class for SWMM errors."""

    def __init__(self, error_code, error_message):
        self.warning = False
        self.args = (error_code,)
        self.message = error_message

    def __str__(self):
        return self.message


class PYSWMMException(Exception):
    """Custom exception class for PySWMM errors."""

    def __init__(self, error_message):
        self.warning = False
        self.message = error_message

    def __str__(self):
        return self.message


class PySWMM(object):
    """
    Wrapper class to lead SWMM DLL object.

    This allow performing operations on the SWMM object that is created when
    the file is being loaded.

    PySWMM can be run in two different modes:

    Mode 1: Execute simulation without any intervention

    * Open the Input file using swmm_open()
    * Execute the simlation without intervening swmmExec()
    * then close calling swmm_close()

    Examples:

    >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
    >>> swmm_model.swmm_open()
    >>> swmm_model.swmmExec()
    >>> swmm_model.swmm_close()

    ---or---

    Mode 2: Step through the entire simulation manually by (This mode allows
    the user to invervene and stream simulation data as well as set parameters
    and use outside controls approaches)

    * swmm_open()
    * swmm_start()
    * swmm_step() or swmm_stride() until it returns 0
    * swmm_end()
    * swmm_report()
    * swmm_close()

    Examples:

    >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
    >>> swmm_model.swmm_open()
    >>> swmm_model.swmm_start(True)
    >>> while(True):
    ...     time = swmm_model.swmm_step() # or swmm_stride()
    ...     if (time <= 0.0): break
    >>>
    >>> swmm_model.swmm_end()
    >>> swmm_model.swmm_report()
    >>> swmm_model.swmm_close()
    """

    def __init__(self, inpfile="", rptfile=None, binfile=None, swmm_lib_path=None):
        """
        Initialize the PySWMM object class.

        User can specified SWMM library path. Uses default lib if
        not provided.

        :param str inpfile: Name of SWMM input file (default '')
        :param str rptfile: Report file to generate (default None)
        :param str binfile: Optional binary output file (default None)
        :param str swmm_lib_path: SWMM library path (default None).

        """
        self.fileLoaded = False
        self.inpfile = inpfile
        self.rptfile = rptfile
        self.binfile = binfile

        self.curSimTime = 0.0

    def swmmExec(self, inpfile=None, rptfile=None, binfile=None):
        """
        Open an input file, run SWMM, then close the file.

        :param str inpfile: Name of SWMM input file
        :param str rptfile: Report file to generate
        :param str binfile: Optional binary output file

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmmExec()
        >>> swmm_model.swmm_close()
        """
        if inpfile is None:
            inpfile = self.inpfile

        if rptfile is None:
            if self.rptfile != "" and self.rptfile is not None:
                rptfile = self.rptfile
            else:
                rptfile = self.inpfile.replace(".inp", ".rpt")

        if binfile is None:
            if self.binfile != "" and self.binfile is not None:
                binfile = self.binfile
            else:
                binfile = self.inpfile.replace(".inp", ".out")

        sys.stdout.write("\n... SWMM Version {}".format(self.swmm_getVersion()))

        try:
            self.swmm_run()
            sys.stdout.write("\n... Run Complete")
        except PYSWMMException:
            sys.stdout.write("\n\n... SWMM completed. There are errors.\n")
            raise (PYSWMMException("SWMM Close Failed"))

    def swmm_run(self, inpfile=None, rptfile=None, binfile=None):
        """# TODO:."""
        if inpfile is None:
            inpfile = self.inpfile

        if rptfile is None:
            if self.rptfile != "" and self.rptfile is not None:
                rptfile = self.rptfile
            else:
                rptfile = self.inpfile.replace(".inp", ".rpt")

        if binfile is None:
            if self.binfile != "" and self.binfile is not None:
                binfile = self.binfile
            else:
                binfile = self.inpfile.replace(".inp", ".out")

        solver.swmm_run(inpfile, rptfile, binfile)

    def swmm_open(self, inpfile=None, rptfile=None, binfile=None):
        """
        Opens SWMM input file & reads in network data.

        :param str inpfile: Name of SWMM input file
        :param str rptfile: Report file to generate
        :param str binfile: Optional binary output file

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_close()
        """
        if self.fileLoaded:
            self.swmm_close()
            error_msg = "Fatal error closing previously opened file"
            raise (PYSWMMException(error_msg))

        if inpfile is None:
            inpfile = self.inpfile

        if rptfile is None:
            if self.rptfile != "" and self.rptfile is not None:
                rptfile = self.rptfile
            else:
                rptfile = self.inpfile.replace(".inp", ".rpt")
                self.rptfile = rptfile

        if binfile is None:
            if self.binfile != "" and self.binfile is not None:
                binfile = self.binfile
            else:
                binfile = self.inpfile.replace(".inp", ".out")
                self.binfile = binfile

        solver.swmm_open(inpfile, rptfile, binfile)
        self.fileLoaded = True

    def swmm_start(self, SaveOut2rpt=False):
        """
        Prepares to Start SWMM Simulation.

        :param bool SaveOut2rpt: Save timeseries results to rpt file
                                 (default is False).

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        solver.swmm_start(SaveOut2rpt)

    def swmm_end(self):
        """
        Ends SWMM Simulation.

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        solver.swmm_end()

    def swmm_step(self):
        """
        Advances SWMM Simulation by a single routing step.

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        return solver.swmm_step()

    def swmm_stride(self, advanceSeconds):
        """
        This function allows for user defined stride length to advance
        the model simulation by a defined time.  This is useful when control
        rules are managed externally by PySWMM. Instead of evaluating rules
        every routing step, instead the simulation can be advanced further
        in time before the PySWMM can intervene. When a 0 is returned, the
        simulation period has reached the end.

        :param int advanceSeconds: Number seconds to advance the simulation
                                   forward.
        :return: Current simulation time after a stride in decimal days (float)
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_stride(600)
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        self.curSimTime = solver.swmm_stride(advanceSeconds)

        return self.curSimTime

    def swmm_report(self):
        """
        Copies Time Series results from .out to .rpt file.

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        solver.swmm_report()

    def swmm_close(self):
        """
        Closes model and supporting files and cleans up memory.

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     if (time <= 0.0): break
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        solver.swmm_close()
        self.fileLoaded = False

    def swmm_save_hotstart(self, hotstart_filename):
        """
        Save the current state of the model to a hotstart file.

        This can be run at any point during the simultion.

        :param str hotstart_filename: Name of hotstart file to save to.
        """
        solver.swmm_hotstart(tka.HotstartFile.save, hotstart_filename)

    def swmm_use_hotstart(self, hotstart_filename):
        """
        Use a hotstart file to initialize the model.

        This must be run before swmm_start() and after swmm_open().

        :param str hotstart_filename: Name of hotstart file to load from.
        """
        solver.swmm_hotstart(tka.HotstartFile.use, hotstart_filename)

    def swmm_getVersion(self):
        """
        Retrieves version number of current SWMM engine.

        The format used is xyzzz where x = major version number,
        y = minor version number, and zzz = build number.

        :return: version number of the DLL source code
        :rtype: int
        """
        major, minor, patch = solver.swmm_version_info().split(".")
        return packaging.version.parse(".".join([major, minor, patch]))

        return

    def swmm_getMassBalErr(self):
        """
        Get Mass Balance Errors.

        :return: Runoff Error, Flow Routing Error, Quality Error
        :rtype: tuple
        """
        return solver.swmm_get_mass_balance()

    # --- NETWORK API FUNCTIONS
    # -------------------------------------------------------------------------
    def getSimulationDateTime(self, timeType):
        """
        Get Simulation Time Data (Based on SimulationTime options).

        :param int timeType: (toolkitapi.SimulationTime member variable)
        :return: datetime
        :rtype: datetime

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getSimulationDateTime(SimulationTime.StartDateTime)
        2015-11-01 14:00:00
        >>> swmm_model.getSimulationDateTime(SimulationTime.EndDateTime)
        2015-11-04 00:00:00
        >>> swmm_model.getSimulationDateTime(SimulationTime.ReportStart)
        2015-11-01 14:00:00
        >>>
        >>> swmm_model.swmm_close()
        """
        return datetime(*solver.simulation_get_datetime(timeType))

    def setSimulationDateTime(self, timeType, newDateTime):
        """
        Set Simulation Time Data (Based on SimulationTime options).

        :param int timeType: (toolkitapi.SimulationTime member variable)

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setSimulationDateTime(SimulationTime.StartDateTime,
                                             datetime(2009, 10, 1, 12,30))
        >>>
        """
        solver.simulation_set_datetime(
            timeType,
            newDateTime.year,
            newDateTime.month,
            newDateTime.day,
            newDateTime.hour,
            newDateTime.minute,
            newDateTime.second,
        )

    def getSimUnit(self, unit_type):
        """
        Get Simulation Units.

        :param int unittype: Simulation Unit Type
        :return: Simulation Unit Type
        :rtype: str

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getSimUnit(SimulationUnits.FlowUnits)
        CFS
        >>> swmm_model.swmm_close()
        """
        value = solver.simulation_get_unit(unit_type)

        if unit_type == tka.SimulationUnits.FlowUnits.value:
            # Temporary Solution (2017-1-2 BEM)
            _flowunitnames = ["CFS", "GPM", "MGD", "CMS", "LPS", "MLD"]
            return _flowunitnames[value]

        elif unit_type == tka.SimulationUnits.UnitSystem.value:
            # Temporary Solution (2017-1-2 BEM)
            _flowunitnames = ["US", "SI"]
            return _flowunitnames[value]

    def getSimOptionSetting(self, setting_type):
        """
        Get Simulation Option Settings.

        :param int settingtype: Analysis Option Setting
        :return: Simulation Analysis option setting
        :rtype: bool

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getSimAnalysisSetting(SimAnalysisSettings.AllowPonding)
        False
        >>> swmm_model.swmm_close()
        """
        return bool(solver.simulation_get_setting(setting_type))

    def getSimAnalysisSetting(self, param_type):
        """
        Get Simulation Configuration Parameter.

        :param int paramtype: Simulation Parameter Type
        :return: Simulation Analysis Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getSimAnalysisSetting(SimulationParameters.RouteStep)
        300
        >>> swmm_model.swmm_close()
        """
        return solver.simulation_get_parameter(param_type)

    def getProjectSize(self, object_type):
        """
        Get Project Size: Number of Objects.

        :param int objecttype: (member variable)
        :return: Object Count
        :rtype: int

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getProjectSize(ObjectType.NODE)
        10
        >>> swmm_model.swmm_close()
        """
        return solver.project_get_count(object_type)

    def getObjectId(self, objecttype, index):
        """
        Get Object ID name.

        :param int objecttype: (member variable)
        :param index: ID Index
        :return: Object ID
        :rtype: string

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getObjectId(ObjectType.NODE,35)
        "example_id_name"
        >>>
        >>> swmm_model.swmm_close()
        """
        return solver.project_get_id(objecttype, index)

    def getObjectIDList(self, objecttype):
        """
        Get Object ID list.

        :param int objecttype: (member variable)

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getObjectIDList(ObjectType.LINK)
        ['C1:C2', 'C2', 'C3']
        >>>
        >>> swmm_model.swmm_close()
        >>>
        """
        IDS = []
        for index in range(self.getProjectSize(objecttype)):
            IDS.append(self.getObjectId(objecttype, index))
        return IDS

    def getObjectIDIndex(self, objecttype, ID):
        """Get Object ID Index. Mostly used as an internal function."""
        return solver.project_get_index(objecttype, ID)

    def ObjectIDexist(self, objecttype, ID):
        """Check if Object ID Exists. Mostly used as an internal function."""
        # Incurred some micro-tech debt. This updated implementation will cover the before and after
        # case of removing this line in SWMM.  Currently the SWMM function throws a non-zero error code. As
        # a result, when it hit swmm-python(swmm-toolkit), it throws an exception.
        # https://github.com/pyswmm/Stormwater-Management-Model/blob/459db1d4dfc61ff994ae01f92eae64e378e08915/src/solver/toolkit.c#L170
        try:
            # eventually this function will return -1 if the index does not exist.
            index = solver.project_get_index(objecttype, ID)
        except:
            index = -1

        if index != -1:
            return True
        else:
            return False

    def getNodeType(self, ID):
        """
        Get Node Type (e.g. Junction, Outfall, Storage, Divider).

        :param str index: ID
        :return: Object ID
        :rtype: int

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getNodeType('J1')
        0
        >>>
        >>> swmm_model.getNodeType('J1') is NodeType.junction
        True
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        return solver.node_get_type(index)

    def getLinkType(self, ID):
        """
        Get Link Type (e.g. Conduit, Pump, Orifice, Weir, Outlet).

        :param str index: ID
        :return: Object ID
        :rtype: int

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLinkType('C1')
        3
        >>>
        >>> swmm_model.getLinkType('C1') is LinkType.weir
        True
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        return solver.link_get_type(index)

    def getLinkConnections(self, ID):
        """
        Get Link Connections (Upstream and Downstream Nodes).

        Interestingly, if the dynamic wave solver is used, when the input file
        is parsed and added to the SWMM5 data model, any negatively sloped
        conduits are reversed automatically. The swmm_getLinkConnections
        function always calls the _swmm_getLinkDirection function to ensure
        the user-assigned upstream ID and downstream IDs are in the correct
        order. This way, the function provides support for directed graphs
        automatically.

        :param str index: ID
        :return: (Upstream Node Index, Downstream Node Index)
        :rtype: tuple

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLinkConnections('C1')
        ('NodeUSID','NodeDSID')
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        us_node, ds_node = solver.link_get_connections(index)
        us_node_id = self.getObjectId(tka.ObjectType.NODE.value, us_node)
        ds_node_id = self.getObjectId(tka.ObjectType.NODE.value, ds_node)

        if self._getLinkDirection(ID) == 1:
            # Return Tuple of Upstream and Downstream Node IDS
            return us_node_id, ds_node_id
        # link validations reverse the conduit direction if the slope is < 0
        elif self._getLinkDirection(ID) == -1:
            # Return Tuple of Upstream and Downstream Node IDS
            return ds_node_id, us_node_id

    def _getLinkDirection(self, ID):
        """
        Internal Method: returns conduit flow direction.

        :param str index: link ID
        :return: 1 for conduit flow from upstream node to downstream node
        and -1 for conduit flow from downstream node to upstream node
        :rtype: int
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        return solver.link_get_direction(index)

    def getNodeParam(self, ID, parameter):
        """
        Get Node Parameter.

        :param str ID: Node ID
        :param int parameter: Parameter (toolkitapi.NodeParams member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getNodeParam('J2',NodeParams.invertElev )
        13.392
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        return solver.node_get_parameter(index, parameter)

    def setNodeParam(self, ID, parameter, value):
        """
        Set Node Parameter.

        :param str ID: Node ID
        :param int Parameter: Parameter (toolkitapi.NodeParams member variable)

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getNodeParam('J2',NodeParams.invertElev, 19 )
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        solver.node_set_parameter(index, parameter, value)

    def getLinkParam(self, ID, parameter):
        """
        Get Link Parameter.

        :param str ID: Link ID
        :param int Parameter: Parameter (toolkitapi.NodeParams member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLinkParam('C1:C2',LinkParams.offset1 )
        0.0
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        return solver.link_get_parameter(index, parameter)

    def setLinkParam(self, ID, parameter, value):
        """
        Set Link Parameter.

        :param str ID: Link ID
        :param int Parameter: Parameter (toolkitapi.NodeParams member variable)

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setLinkParam('C1:C2',LinkParams.offset1, 2 )
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        solver.link_set_parameter(index, parameter, value)

    def getLidCOverflow(self, ID):
        """
        Get Lid Control Parameter.

        :param str ID: Lid Control ID
        :rtype: Bool

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLidCOverflow('J2')
        True
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LID.value, ID)
        return bool(solver.lid_control_get_overflow(index))

    def getLidCParam(self, ID, layer, parameter):
        """
        Get LidControl Parameter.

        :param str ID: Lid Control ID
        :param int layer: Layer (toolkitapi.LidLayers member variable)
        :param int parameter: Parameter (toolkitapi.LidLayersProperty member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLidCParam('J2', LidLayer.surface, LidLayersProperty.thickness)
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LID.value, ID)
        return solver.lid_control_get_parameter(index, layer, parameter)

    def setLidCParam(self, ID, layer, parameter, value):
        """
        Set Lid Control Parameter Before/During Model Simulation.

        :param str ID: Lid Control ID
        :param int layer: Layer (toolkitapi.LidLayers member variable)
        :param int parameter: Parameter (toolkitapi.LidLayersProperty member variable)

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setLidCParam('J2', LidLayer.surface, LidLayersProperty.thickness, 110)
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LID.value, ID)
        solver.lid_control_set_parameter(index, layer, parameter, value)

    def getLidUCount(self, ID):
        """
        Get Number of Lid Units Defined for Subcatchment.

        :param str ID: Subcatchment ID
        :rtype: int

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLidUCount('J2')
        2
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        return solver.lid_usage_get_count(index)

    def getLidUParam(self, subcatchID, lid_index, parameter):
        """
        Get LidUnit Parameter

        :param str subcatchID: Subcatchment ID
        :param int lid_index: Lid unit index
        :param int parameter: Paramter (toolkitapi.LidUParams member variable)
        :return: paramater Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLidUParam('S2', 0, LidUParams.unitarea)
        1000
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        return solver.lid_usage_get_parameter(index, lid_index, parameter)

    def setLidUParam(self, subcatchID, lid_index, parameter, value):
        """
        Set LidUnit Parameter

        :param str subcatchID: Subcatchment ID
        :param int lid_index: Lid unit index
        :param int parameter: parameter (toolkitapi.LidUParams member variable)
        :param double value: value set to parameter
        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setLidUParam('S2', 0, LidUParams.unitarea, 10)
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        solver.lid_usage_set_parameter(index, lid_index, parameter, value)

    def getLidUOption(self, subcatchID, lid_index, parameter):
        """
        Get LidUnit Option

        :param str subcatchID: Subcatchment ID
        :param int lid_index: Lid unit Index
        :param int parameter: paramter (toolkitapi.LidUParams member variable)
        :return: paramater Value
        :rtype: int

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLidUOption('S2', 0, LidUParams.index)
        1000
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        return solver.lid_usage_get_option(index, lid_index, parameter)

    def setLidUOption(self, subcatchID, lid_index, parameter, value):
        """
        Set LidUnit Option

        :param str subcatchID: Subcatchment ID
        :param int lid_index: Lid unit index
        :param int parameter: paramter (toolkitapi.LidUParams member variable)
        :param double value: value set to parameter
        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setLidUOption('S2', 0, LidUParams.index, 0)
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        solver.lid_usage_set_option(index, lid_index, parameter, value)

    def getSubcatchParam(self, ID, parameter):
        """
        Get Subcatchment Parameter

        :param str ID: Subcatchment ID
        :param int Parameter: Parameter (toolkitapi.SubcParams member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getLinkParam('S2',SubcParams.area )
        43561.596096880996
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        return solver.subcatch_get_parameter(index, parameter)

    def setSubcatchParam(self, ID, parameter, value):
        """
        Set Subcatchment Parameter.

        :param str ID: Subcatchment ID
        :param int parameter: paramter (toolkitapi.SubcParams member variable)
        :param double value: value set to parameter

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setLinkParam('S2',SubcParams.area, 100 )
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        solver.subcatch_set_parameter(index, parameter, value)

    def getSubcatchOutConnection(self, ID):
        """
        Get Subcatchment Outlet Connection.

        This function return the type of loading surface and the ID. The two
        load to objects are nodes and other subcatchments.

        Nodes are ObjectType.NODE
        Subcatchments are ObjectType.SUBCATCH

        :param str ID: Subcatchment ID
        :param int Parameter: Parameter (toolkitapi.SubcParams member variable)
        :return: (Loading Surface Type, ID)
        :rtype: tuple

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getSubcatchOutConnection('S2',SubcParams.area )
        (2, 'J2')
        >>>
        >>> subout = swmm_model.getSubcatchOutConnection('S2',SubcParams.area )
        >>> subout[0]  == ObjectType.NODE
        True
        >>>
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        outlet_type, outlet_index = solver.subcatch_get_connection(index)
        outlet_id = self.getObjectId(outlet_type, outlet_index)

        return outlet_id

    def getGagePrecip(self, ID, parameter):
        """
        Get precipitation from gage

        This function returns the rainfall, show and total precipitation
        associated with the gage

        :param str ID: Gage ID
        :param int parameter: paramter (toolkitapi.RainGageResults member variable)
        :return: value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.getGagePrecip('Gage1', )
        0.0
        >>> swmm_model.swmm_close()

        """
        index = self.getObjectIDIndex(tka.ObjectType.GAGE.value, ID)
        return solver.raingage_get_precipitation(index, parameter)

    # --- Active Simulation Result "Getters"
    # -------------------------------------------------------------------------
    def getCurrentSimulationTime(self):
        """
        Get Current Simulation DateTime in Python Format.

        :return: Python Datetime
        :rtype: datetime

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getCurrentSimualationTime())
        ...     if (time <= 0.0): break
        2015-11-03 10:10:12
        2015-11-03 10:20:12
        2015-11-03 10:30:12
        2015-11-03 10:40:12
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        return datetime(*solver.simulation_get_current_datetime())

    def getLidUFluxRates(self, subcatchID, lidIndex, layerIndex):
        """
        Get Lid Unit Layer Flux Rates Result.

        :param str subcatchID: Subcatchment ID
        :param int lidIndex: Lid unit Index
        :param int layerIndex: Parameter (toolkitapi.LidLayers member variable)
        :return: Parameter Value
        :rtype: float
        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getLidUFluxRates('J1', 0, LidLayers.surface))
        ...     if (time <= 0.0): break
        1.2
        1.5
        1.9
        1.2
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        return solver.lid_usage_get_flux_rate(index, lidIndex, layerIndex)

    def getLidUResult(self, subcatchID, lidIndex, resultType):
        """
        Get Lid Result.

        :param str subcatchID: Subcatchment ID
        :param int lidIndex: Lid unit Index
        :param int resultType: Parameter (toolkitapi.LidUResults member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getLidUResult('J1', 0, LidUResults.inflow))
        ...     if (time <= 0.0): break
        1.2
        1.5
        1.9
        1.2
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        return solver.lid_usage_get_result(index, lidIndex, resultType)

    def getLidGResult(self, subcatchID, resultType):
        """
        Get Lid Group Result.

        :param str subcatchID: Subcatchment ID
        :param int resultType: Parameter (toolkitapi.LidUResults member variable)
        :return: Parameter Value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getLidGResult('J1', LidUResults.flowToPerv))
        ...     if (time <= 0.0): break
        1.2
        1.5
        1.9
        1.2
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, subcatchID)
        return solver.lid_group_get_result(index, resultType)

    def getNodeResult(self, ID, result_type):
        """
        Get Node Result.

        :param str ID: Node ID
        :param int result_type: parameter (toolkitapi.NodeResults member variable)
        :return: value
        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getNodeResult('J1', NodeResults.newDepth))
        ...     if (time <= 0.0): break
        1.2
        1.5
        1.9
        1.2
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        return solver.node_get_result(index, result_type)

    def getNodePollut(self, ID, result_type):
        """
        Get water quality results from a Node.

        :param str ID: Node ID
        :param int result_type: parameter (toolkitapi.NodePollut member value)
        :rtype: list
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        return solver.node_get_pollutant(index, result_type)

    def getLinkResult(self, ID, result_type):
        """
        Get Link Result.

        :param str ID: Link ID
        :param int result_type: parameter (toolkitapi.LinkResults member variable)
        :return: parameter value

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getLinkResult('J1', LinkResults.newFlow))
        ...     if (time <= 0.0): break
        1.2
        1.5
        1.9
        1.2
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        return solver.link_get_result(index, result_type)

    def getLinkPollut(self, ID, result_type):
        """
        Get water quality results from a Link.

        :param str ID: Link ID
        :param int result_type:  parameter (toolkitapi.LinkPollut member value)
        :return: Pollutant Value
        :rtype: list
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        return solver.link_get_pollutant(index, result_type)

    def getSubcatchResult(self, ID, result_type):
        """
        Get Subcatchment Result

        :param str ID: Subcatchment ID
        :param int result_type: paramter (toolkitapi.LinkResults member variable)
        :return: paramater Value

        :rtype: float

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     print(swmm_model.getSubcatchResult('S3', SubcResults.newRunoff))
        ...     if (time <= 0.0): break
        0.01
        0.05
        0.09
        0.08
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        return solver.subcatch_get_result(index, result_type)

    def getSubcatchPollut(self, ID, result_type):
        """
        Get pollutant results from a Subcatchment.

        :param str ID: Subcatchment ID
        :param int result_type: parameter (toolkitapi.SubcPollut member variable)
        :return: Pollutant Values
        :rtype: list
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        return solver.subcatch_get_pollutant(index, result_type)

    def node_statistics(self, ID):
        """
        Get stats for a Node.

        :param str ID: Node ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        stats = solver.node_get_stats(index)
        alias = tka.NodeStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def node_inflow(self, ID):
        """
        Get total inflow volume for a Node.

        :param str ID: Node ID
        :return: Total Volume
        :rtype: float
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        return solver.node_get_total_inflow(index)

    def storage_statistics(self, ID):
        """
        Get stats for a Storage Node.

        :param str ID: Node ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        stats = solver.storage_get_stats(index)
        alias = tka.StorageStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def outfall_statistics(self, ID):
        """
        Get stats for a Outfall Node.

        :param str ID: Node ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        stats = solver.outfall_get_stats(index)
        alias = tka.OutfallStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                if attr == "totalLoad":
                    dict_stats[alias[attr]] = {}
                    pollutants = self.getObjectIDList(tka.ObjectType.POLLUT.value)
                    for index, pollutant in enumerate(pollutants):
                        dict_stats[alias[attr]][pollutant] = stats.get_totalLoad(index)
                else:
                    dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def conduit_statistics(self, ID):
        """
        Get stats for a Conduits.

        :param str ID: Conduit ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        stats = solver.link_get_stats(index)
        alias = tka.LinkStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def pump_statistics(self, ID):
        """
        Get stats for a Pump.

        :param str ID: Pump ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        stats = solver.pump_get_stats(index)
        alias = tka.PumpStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def subcatch_statistics(self, ID):
        """
        Get stats for a Subcatchment.

        :param str ID: Subcatchment ID
        :return: Group Stats
        :rtype: dict
        """
        index = self.getObjectIDIndex(tka.ObjectType.SUBCATCH.value, ID)
        stats = solver.subcatch_get_stats(index)
        alias = tka.SubcStats._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def flow_routing_stats(self):
        """
        Get Flow Routing System stats.

        :return: Dictionary of Flow Routing Stats.
        :rtype: dict
        """
        stats = solver.system_get_routing_totals()
        alias = tka.RoutingTotals._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    def runoff_routing_stats(self):
        """
        Get Runoff Routing System stats.

        :return: Dictionary of Runoff Routing Stats.
        :rtype: dict
        """
        stats = solver.system_get_runoff_totals()
        alias = tka.RunoffTotals._py_alias_ids
        # Copy Items to Dictionary using Alias Names.
        dict_stats = {}
        for attr in dir(stats):
            if "_" not in attr and attr in alias:
                dict_stats[alias[attr]] = getattr(stats, attr)
        return dict_stats

    # --- Active Simulation Parameter "Setters"
    # -------------------------------------------------------------------------
    def setLinkSetting(self, ID, target_setting):
        """
        Set Link Setting (Pumps, Orifices, Weirs).

        :param str ID: Link ID
        :param float target_setting: New target setting which will be applied
        at the start of the next routing step.

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> i = 0
        >>> while(True):
        ...     time = swmm_model.swmm_step()
        ...     i+=1
        ...     if i == 80:
        ...         swmm_model.setLinkSetting('C3',0.5)
        ...     if (time <= 0.0): break
        ...
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        solver.link_set_target_setting(index, target_setting)

    def setNodeInflow(self, ID, flow_rate):
        """
        Set Node Inflow rate.

        The flow rate should be in the user defined units. The value is held
        constant in the model until it is redefined by the toolkit API.

        :param str ID: Node ID
        :param float flow_rate: New flow rate in the user-defined flow units

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> i = 0
        >>> while(True):
        ...     if i == 80:
        ...         swmm_model.setNodeInflow('J1',4)
        ...     time = swmm_model.swmm_step()
        ...     i+=1
        ...     if (time <= 0.0): break
        ...
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        solver.node_set_total_inflow(index, flow_rate)

    def setOutfallStage(self, ID, stage):
        """
        Set Outfall Stage (head).

        The level should be in the user defined units. The value is held
        constant in the model until it is redefined by the toolkit API.

        :param str ID: Node ID
        :param float stage: New flow rate in the user-defined flow units

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.swmm_start(True)
        >>> i = 0
        >>> while(True):
        ...     if i == 80:
        ...         swmm_model.setOutfallStage('J1',4)
        ...     time = swmm_model.swmm_step()
        ...     i+=1
        ...     if (time <= 0.0): break
        ...
        >>>
        >>> swmm_model.swmm_end()
        >>> swmm_model.swmm_report()
        >>> swmm_model.swmm_close()
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        solver.outfall_set_stage(index, stage)

    def setGagePrecip(self, ID, value):
        """
        Set precipitation to gage

        This function sets the rainfall intensity to the gage

        :param str ID: Gage ID
        :param float valve: rainfall intensity
        :return: errcode

        Examples:

        >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
        >>> swmm_model.swmm_open()
        >>> swmm_model.setGagePrecip('Gage1', 10.0)
        >>> swmm_model.swmm_close()

        """
        index = self.getObjectIDIndex(tka.ObjectType.GAGE.value, ID)
        solver.raingage_set_precipitation(index, value)

    def setNodePollut(self, ID, pollutant_ID, pollutant_value):
        """
        Set water quality results in a Node.

        :param str ID: Node ID
        :param str pollutant_ID
        :param float pollutant_value: pollutant value to set
        """
        index = self.getObjectIDIndex(tka.ObjectType.NODE.value, ID)
        pollutant_index = self.getObjectIDIndex(
            tka.ObjectType.POLLUT.value, pollutant_ID
        )
        solver.node_set_pollutant(
            index, tka.NodePollut.nodeQual.value, pollutant_index, pollutant_value
        )

    def setLinkPollut(self, ID, pollutant_ID, pollutant_value):
        """
        Set water quality results in a Link.

        :param str ID: Link ID
        :param str ID: Pollutant ID
        :param float pollutant_value: pollutant value to set
        """
        index = self.getObjectIDIndex(tka.ObjectType.LINK.value, ID)
        pollutant_index = self.getObjectIDIndex(
            tka.ObjectType.POLLUT.value, pollutant_ID
        )
        solver.link_set_pollutant(
            index, tka.LinkPollut.linkQual.value, pollutant_index, pollutant_value
        )


if __name__ == "__main__":
    test = PySWMM(
        inpfile=r"./tests/data/model_weir_setting.inp",
        rptfile=r"./tests/data/model_weir_setting.rpt",
        binfile=r"./tests/data/model_weir_setting.out",
    )
    test.swmm_open()

    print("Simulation Time Info")
    print("Start Time")
    print(test.getSimulationDateTime(tka.SimulationTime.StartDateTime.value))
    print("End Time")
    print(test.getSimulationDateTime(tka.SimulationTime.EndDateTime.value))
    print("Report Time")
    print(test.getSimulationDateTime(tka.SimulationTime.ReportStart.value))

    print("Simulation Units")
    print(test.getSimUnit(tka.SimulationUnits.FlowUnits.value))

    print("Simulation Engine Version")
    print(test.swmm_getVersion())

    print("Simulation Allow Ponding Option Selection")
    print(
        test.getSimAnalysisSetting(tka.SimAnalysisSettings.AllowPonding.value),
    )

    print("Simulation Routing Step")
    print(test.getSimAnalysisSetting(tka.SimulationParameters.RouteStep.value))

    print("Number of Nodes")
    print(test.getProjectSize(tka.ObjectType.NODE.value))

    print("Node ID")
    IDS = test.getObjectIDList(tka.ObjectType.NODE.value)
    print(IDS)
    print("ID,Invert,Type")
    for ind, idd in enumerate(IDS):
        print(
            ind,
            idd,
            test.getNodeParam(idd, tka.NodeParams.invertElev.value),
            test.getNodeParam(idd, tka.NodeParams.fullDepth.value),
            test.getNodeType(idd),
        )

    print("Link ID")
    print("ID,offset1,LinkConnections")
    IDS = test.getObjectIDList(tka.ObjectType.LINK.value)
    print(IDS)
    for ind, idd in enumerate(IDS):
        print(
            ind,
            idd,
            test.getLinkParam(idd, tka.LinkParams.offset1.value),
            test.getLinkConnections(idd),
        )

    print("SUBCATCH ID")
    IDS = test.getObjectIDList(tka.ObjectType.SUBCATCH.value)
    print(IDS)
    for ind, idd in enumerate(IDS):
        print(
            ind,
            idd,
            test.getSubcatchParam(idd, tka.SubcParams.area.value),
            test.getSubcatchOutConnection(idd),
        )

    test.swmm_start(False)
    i = 0
    while True:
        if i % 1000 == 0:
            print("test {}".format(test.flow_routing_stats()))
        eltime = test.swmm_step()
        i += 1
        if eltime == 0:
            break

    test.swmm_end()
    test.swmm_close()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2018 Jennifer Wu
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from pyswmm.swmm5 import PYSWMMException
from pyswmm.toolkitapi import ObjectType
from pyswmm.lidlayers import Surface, Soil, Storage, Pavement, Drain, DrainMat


class LidControls(object):
    """Lid Control Iterator Methods.

    :param object model: Open Model Instance

    This allows the user to iterate and get on of the LID control types as
    defined inside SWMM's [LID_CONTROLS] section.  For example, here is a sample
    section from a SWMM INP file.

    .. code-block::

        [LID_CONTROLS]
        ;;Name           Type_Layer Parameters
        ;;-------------- ---------- ----------
        LID_C1  IT
        LID_C1  SURFACE    0.0    0.0    0.03   1.0    5
        LID_C1  STORAGE    800    0.3    3.6    0      NO
        LID_C1  DRAIN      0.5    1.41   400    6      0       0

    Now we can peform manipulations in PySWMM and get objects.

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            for lid_control in LidControls(sim):
                print(lid_control)

    .. code-block::

        >>> LID_C1
        >>> lid_control_type_2

    It is also possible to simply get an LID control by name as follows:

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']
            print(lid_control)

    .. code-block::

        >>> LID_C1

    Once the LID control object is instantiated, there are getter and setters
    available for each different layer of the LID Control.  The layers are the
    following and these become instance attributes of each LID Control (LidControl)
    object.

    +------------+-----------+
    | Layer Type | Attribute |
    +============+===========+
    | Surface    | surface   |
    +------------+-----------+
    | Soil       | soil      |
    +------------+-----------+
    | Storage    | storage   |
    +------------+-----------+
    | Pavement   | pavement  |
    +------------+-----------+
    | Drain      | drain     |
    +------------+-----------+
    | DrainMat   | drain_mat |
    +------------+-----------+

    In our example model section above the layers defined are Surface,
    Storage, and Drain.  Once the handles are available, the properties
    and setter methods can be called to read and modify the values.

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_surface = lid_control.surface
            lid_control_storage = lid_control.storage
            lid_control_drain = lid_control.drain

            print(lid_control_surface.roughness)
            lid_control_surface.roughness = 0.5
            print(lid_control_surface.roughness)

            for step in sim:
                pass

    .. code-block::

        >>> 0.03
        >>> 0.5

    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._sim = model
        self._model = model._model
        self._cuindex = 0
        self._nlidcontrols = self._model.getProjectSize(ObjectType.LID.value)

    def __len__(self):
        """
        Return number of lid controls.

        Use the expression 'len(LidControls)'.

        :return: Number of Lid Controls
        :rtype: int

        """
        return self._model.getProjectSize(ObjectType.LID.value)

    def __contains__(self, lidcontrolid):
        """
        Checks if Lid Control ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.LID.value, lidcontrolid)

    def __getitem__(self, lidcontrolid):
        if self.__contains__(lidcontrolid):
            return LidControl(self._sim, self._model, lidcontrolid)
        else:
            raise PYSWMMException("Lid Control ID Does not Exist")

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nlidcontrols:
            lidcontrolobject = self.__getitem__(self._lidcontrolid)
            self._cuindex += 1  # Next Iteration
            return lidcontrolobject
        else:
            raise StopIteration()

    @property
    def _lidcontrolid(self):
        """Lid Control ID."""
        return self._model.getObjectId(ObjectType.LID.value, self._cuindex)


class LidControl(object):
    """
    Once the LID control object is instantiated, there are getter and setters
    available for each different layer of the LID Control.  The layers are the
    following and these become instance attributes of each LID Control (LidControl)
    object.

    +------------+-----------+
    | Layer Type | Attribute |
    +============+===========+
    | Surface    | surface   |
    +------------+-----------+
    | Soil       | soil      |
    +------------+-----------+
    | Storage    | storage   |
    +------------+-----------+
    | Pavement   | pavement  |
    +------------+-----------+
    | Drain      | drain     |
    +------------+-----------+
    | DrainMat   | drain_mat |
    +------------+-----------+
    """

    def __init__(self, sim, model, lidcontrolid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if lidcontrolid not in model.getObjectIDList(ObjectType.LID.value):
            raise PYSWMMException("ID Not valid")
        self._sim = sim
        self._model = model
        self._lidcontrolid = lidcontrolid

        self.surface = Surface(model, self)
        self.soil = Soil(model, self)
        self.storage = Storage(model, self)
        self.pavement = Pavement(model, self)
        self.drain = Drain(model, self)
        self.drain_mat = DrainMat(model, self)

    def __str__(self):
        return self._lidcontrolid

    @property
    def can_overflow(self):
        """
        Get lid control surface layer option for immediate outflow of excess water

        :return: Parameter Value
        :rtype: char
        """
        return self._model.getLidCOverflow(self._lidcontrolid)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Links module for the pythonic interface to SWMM5."""
from swmm.toolkit import shared_enum

# Local imports
from pyswmm.swmm5 import PYSWMMException
from pyswmm.toolkitapi import LinkParams, LinkResults, LinkPollut, LinkType, ObjectType


class Links(object):
    """
    Link Iterator Methods.

    :param object model: Open Model Instance

    Examples:

    >>> from pyswmm import Simulation, Links
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     for link in Links(sim):
    ...         print(link)
    ...         print(link.linkid)
    <swmm5.Link object at 0x031B0350>
    C1
    <swmm5.Link object at 0x030693D0>
    C2
    <swmm5.Link object at 0x031B0350>
    C3
    <swmm5.Link object at 0x030693D0>
    C0

    Iterating or Links Object

    >>> links = Links(sim)
    >>> for link in links:
    ...     print(link.linkid)
    C1:C2
    C2
    C3

    Testing Existence

    >>> links = Links(sim)
    >>> "C1:C2" in links
    True

    Initializing a link Object

    >>> links = Links(sim)
    >>> c1c2 = links['C1:C2']
    >>> c1c2.flow_limit = 12
    >>> c1c2.flow_limit
    12
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model
        self._cuindex = 0
        self._nLinks = self._model.getProjectSize(ObjectType.LINK.value)

    def __len__(self):
        """
        Return number of links. Use the expression 'len(Links)'.

        :return: Number of Links
        :rtype: int
        """
        return self._model.getProjectSize(ObjectType.LINK.value)

    def __contains__(self, linkid):
        """
        Checks if Link ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.LINK.value, linkid)

    def __getitem__(self, linkid):
        if self.__contains__(linkid):
            ln = Link(self._model, linkid)
            _ln = ln
            if ln.is_conduit():
                _ln.__class__ = Conduit
            elif ln.is_pump():
                _ln.__class__ = Pump
            return _ln
        else:
            raise PYSWMMException("Link ID: {} Does not Exist".format(linkid))

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nLinks:
            linkobject = self.__getitem__(self._linkid)
            self._cuindex += 1  # Next Iteration
            return linkobject
        else:
            raise StopIteration()

    @property
    def _linkid(self):
        """Link ID."""
        return self._model.getObjectId(ObjectType.LINK.value, self._cuindex)


class Link(object):
    """
    Link Methods.

    :param object model: Open Model Instance
    :param str linkid: Link ID

    Examples:

    >>> from pyswmm import Simulation, Links
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     c1c2 = Links(sim)["C1:C2"]
    ...     print(c1c2.flow)
    ...     for step in sim:
    ...         print(c1c2.flow)
    0.0
    """

    def __init__(self, model, linkid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if linkid not in model.getObjectIDList(ObjectType.LINK.value):
            raise PYSWMMException("ID Not valid")
        self._model = model
        self._linkid = linkid

    # --- Get Parameters
    # -------------------------------------------------------------------------
    @property
    def linkid(self):
        """
        Get Link ID.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.linkid)
        "C1"
        """
        return self._linkid

    def is_conduit(self):
        """
        Check if link is a Conduit Type.

        :return: is conduit
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.is_conduit())
        True
        """
        return self._model.getLinkType(self._linkid) is shared_enum.LinkType.CONDUIT

    def is_pump(self):
        """
        Check if link is a Pump Type.

        :return: is pump
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.is_pump())
        False
        """
        return self._model.getLinkType(self._linkid) is shared_enum.LinkType.PUMP

    def is_orifice(self):
        """
        Check if link is a Orifice Type.

        :return: is orifie
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.is_orifice())
        False
        """
        return self._model.getLinkType(self._linkid) is shared_enum.LinkType.ORIFICE

    def is_weir(self):
        """
        Check if link is a Weir Type.

        :return: is weir
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.is_weir())
        False
        """
        return self._model.getLinkType(self._linkid) is shared_enum.LinkType.WEIR

    def is_outlet(self):
        """
        Check if link is a Outlet Type.

        :return: is outlet
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.is_outlet())
        False
        """
        return self._model.getLinkType(self._linkid) is shared_enum.LinkType.OUTLET

    @property
    def connections(self):
        """
        Get link upstream and downstream node IDs.

        :return: ("UpstreamNodeID","DownstreamNodeID")
        :rtype: tuple

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.connections)
        ("C1","C2")
        """
        return self._model.getLinkConnections(self._linkid)

    @property
    def inlet_node(self):
        """
        Get link inlet node ID.

        :return: Inlet node ID
        :rtype: str

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.inlet_node)
        C1
        """
        return self._model.getLinkConnections(self._linkid)[0]

    @property
    def outlet_node(self):
        """
        Get link outlet node ID.

        :return: Outlet node ID
        :rtype: str

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.outlet_node)
        C2
        """
        return self._model.getLinkConnections(self._linkid)[1]

    @property
    def inlet_offset(self):
        """
        Get/set Upstream Offset Depth.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.inlet_offset)
        0.1

        Setting the value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.inlet_offset)
        ...     c1c2.inlet_offset = 0.2
        ...     print(c1c2.inlet_offset)
        0.1
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.offset1.value)

    @inlet_offset.setter
    def inlet_offset(self, param):
        """Set Link Upstream Link Offset."""
        self._model.setLinkParam(self._linkid, LinkParams.offset1.value, param)

    @property
    def outlet_offset(self):
        """
        Get/set Downstream Offset Depth.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.outlet_offset)
        0.1

        Setting the value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.outlet_offset)
        ...     c1c2.outlet_offset = 0.2
        ...     print(c1c2.outlet_offset)
        0.1
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.offset2.value)

    @outlet_offset.setter
    def outlet_offset(self, param):
        """Set Link Downstream Link Offset."""
        self._model.setLinkParam(self._linkid, LinkParams.offset2.value, param)

    @property
    def initial_flow(self):
        """
        Get/set Link Initial Flow.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.initial_flow)
        0

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.initial_flow)
        ...     c1c2.initial_flow = 0.2
        ...     print(c1c2.initial_flow)
        0.1
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.q0.value)

    @initial_flow.setter
    def initial_flow(self, param):
        """Set Link Initial Flow Rate."""
        self._model.setLinkParam(self._linkid, LinkParams.q0.value, param)

    @property
    def flow_limit(self):
        """
        Get/set link flow limit.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.flow_limit)
        0

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.flow_limit)
        ...     c1c2.flow_limit = 0.2
        ...     print(c1c2.flow_limit)
        0
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.qLimit.value)

    @flow_limit.setter
    def flow_limit(self, param):
        """Set Link Flow Limit."""
        self._model.setLinkParam(self._linkid, LinkParams.qLimit.value, param)

    @property
    def inlet_head_loss(self):
        """
        Get/set Inlet Head Loss.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.inlet_head_loss)
        0

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.inlet_head_loss)
        ...     c1c2.inlet_head_loss = 0.2
        ...     print(c1c2.inlet_head_loss)
        0
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.cLossInlet.value)

    @inlet_head_loss.setter
    def inlet_head_loss(self, param):
        """Set Link Inlet Head Loss."""
        self._model.setLinkParam(self._linkid, LinkParams.cLossInlet.value, param)

    @property
    def outlet_head_loss(self):
        """
        Get/set Outlet Head Loss.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.outlet_head_loss)
        0

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.outlet_head_loss)
        ...     c1c2.outlet_head_loss = 0.2
        ...     print(c1c2.outlet_head_loss)
        0
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.cLossOutlet.value)

    @outlet_head_loss.setter
    def outlet_head_loss(self, param):
        """Set Link Outlet Head Loss."""
        self._model.setLinkParam(self._linkid, LinkParams.cLossOutlet.value, param)

    @property
    def average_head_loss(self):
        """
        Get/set Average Conduit Loss.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.average_head_loss)
        0

        Setting the value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.average_head_loss)
        ...     c1c2.average_head_loss = 0.2
        ...     print(c1c2.average_head_loss)
        0
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.cLossAvg.value)

    @average_head_loss.setter
    def average_head_loss(self, param):
        """Set Link Average Head Loss."""
        self._model.setLinkParam(self._linkid, LinkParams.cLossAvg.value, param)

    @property
    def seepage_rate(self):
        """
        Get/set Conduit Seepage Loss.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.seepage_rate)
        0

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.seepage_rate)
        ...     c1c2.seepagerate = 0.2
        ...     print(c1c2.seepage_rate)
        0
        0.2
        """
        return self._model.getLinkParam(self._linkid, LinkParams.seepRate.value)

    @seepage_rate.setter
    def seepage_rate(self, param):
        """Set Link Average Seepage Loss."""
        self._model.setLinkParam(self._linkid, LinkParams.seepRate.value, param)

    @property
    def flow(self):
        """
        Get Link Results for Flow.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.flow)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.newFlow.value)

    @property
    def depth(self):
        """
        Get Link Results for Depth.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.depth)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.newDepth.value)

    @property
    def volume(self):
        """
        Get Link Results for Volume.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.volume)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.newVolume.value)

    @property
    def froude(self):
        """
        Get Link Results for Froude.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.froude)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.froude.value)

    @property
    def ups_xsection_area(self):
        """
        Get Link Results for Upstream X-section Flow Area.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.ups_xsection_area)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.surfArea1.value)

    @property
    def ds_xsection_area(self):
        """
        Get Link Results for Downstream X-section Flow Area.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.ds_xsection_area)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getLinkResult(self._linkid, LinkResults.surfArea2.value)

    @property
    def current_setting(self):
        """
        Get Link current setting.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.current_setting)
        0
        1
        0
        0.5
        1
        """
        return self._model.getLinkResult(self._linkid, LinkResults.setting.value)

    @property
    def target_setting(self):
        """
        Get/set Link Target Setting.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.target_setting)
        0
        0
        1
        0.5
        1

        Setting the Value

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     for step in sim:
        ...         print(c1c2.target_setting)
        ...         if c1c2.flow > 3:
        ...             c1c2.target_setting = 0.1
        0
        0
        0.1
        0.1
        0.1
        """
        return self._model.getLinkResult(self._linkid, LinkResults.targetSetting.value)

    @target_setting.setter
    def target_setting(self, setting):
        """
        Set Link Target Setting.

        If Simulation is not running this method will raise a warning and
        return 0.
        """
        return self._model.setLinkSetting(self._linkid, setting)

    @property
    def pollut_quality(self):
        """
        Get Current Water Quality Values for a Link.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_pollutants.inp') as sim:
        ...     C1 = Links(sim)["C1"]
        ...     for step in sim:
        ...         print(C1.pollut_quality)
        {'test-pollutant': 0.0}
        {'test-pollutant': 120.0}
        {'test-pollutant': 120.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        quality_array = self._model.getLinkPollut(
            self._linkid, LinkPollut.linkQual.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = quality_array[ind]

        return out_dict

    @pollut_quality.setter
    def pollut_quality(self, args):
        """
        Set Current Link Water Quality Value

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_pollutants_setters.inp') as sim:
        ...     C1 = Links(sim)["C1"]
        ...     for step in sim:
        ...         C1.pollut_quality = ('test-pollutant', 100)
                    print(C1.pollut_quality)
        {'test-pollutant': 100.0}
        {'test-pollutant': 100.0}
        {'test-pollutant': 100.0}
        """

        pollutant_ID, pollutant_value = args

        self._model.setLinkPollut(self._linkid, pollutant_ID, pollutant_value)

    @property
    def total_loading(self):
        """
        Get Total Pollutant Loading Values for a Link.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Total Loading Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_pollutants.inp') as sim:
        ...     C1 = Links(sim)["C1"]
        ...     for step in sim:
        ...         print(C1.total_loading)
        {'test-pollutant': 0.01}
        {'test-pollutant': 0.02}
        {'test-pollutant': 0.03}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        totalLoad_array = self._model.getLinkPollut(
            self._linkid, LinkPollut.totalLoad.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = totalLoad_array[ind]

        return out_dict

    @property
    def reactor_quality(self):
        """
        Get Current Water Quality Values for the mixed reactor inside a Link.
        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Links
        >>>
        >>> with Simulation('tests/data/model_pollutants.inp') as sim:
        ...     C1 = Links(sim)["C1"]
        ...     for step in sim:
        ...         print(C1.reactor_quality)
        {'test-pollutant': 0.0}
        {'test-pollutant': 120.0}
        {'test-pollutant': 120.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        quality_array = self._model.getLinkPollut(
            self._linkid, LinkPollut.linkQual.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = quality_array[ind]

        return out_dict


class Conduit(Link):
    """
    Conduit Object: Subclass of Link Object.
    """

    def __init__(self):
        super(Conduit, self).__init__()

    @property
    def conduit_statistics(self):
        """
        Conduit Flow Stats. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +-----------------------+
        | peak_flow             |
        +-----------------------+
        | peak_flow_date        |
        +-----------------------+
        | peak_velocity         |
        +-----------------------+
        | peak_depth            |
        +-----------------------+
        | time_normal_flow      |
        +-----------------------+
        | time_inlet_control    |
        +-----------------------+
        | time_surcharged       |
        +-----------------------+
        | time_full_upstream    |
        +-----------------------+
        | time_full_downstream  |
        +-----------------------+
        | time_full_flow        |
        +-----------------------+
        | time_capacity_limited |
        +-----------------------+
        | time_in_flow_class    |
        +-----------------------+
        | time_courant_crit     |
        +-----------------------+
        | flow_turns            |
        +-----------------------+
        | flow_turn_sign        |
        +-----------------------+

        Time in Flow Class: (Fraction of Total Time)

        +-----+--------+----------+----------+----------+---------+-----------+
        | 0   | 1      | 2        | 3        | 4        | 5       | 6         |
        +-----+--------+----------+----------+----------+---------+-----------+
        | Dry | Up Dry | Down Dry | Sub Crit | Sup Crit | Up Crit | Down Crit |
        +-----+--------+----------+----------+----------+---------+-----------+

        :return: Group of Stats
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Links
        >>> import pprint
        >>> pp = pprint.PrettyPrinter(indent=4)
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     c1c2 = Links(sim)["C1:C2"]
        ...     print(c1c2.flow)
        ...     for step in sim:
        ...         pass
        ...     pp.pprint(c1c2.conduit_statistics)
        o  Retrieving project data
        {   'flow_turn_sign': -1,
            'flow_turns': 1,
            'peak_depth': 0.8744979811226141,
            'peak_flow': 9.612897240188659,
            'peak_flow_date': 42310.00000579861,
            'peak_velocity': 13.200109870715737,
            'time_capacity_limited': 0.0,
            'time_courant_crit': 0.0,
            'time_full_downstream': 0.0,
            'time_full_flow': 57.98847222222222,
            'time_full_upstream': 57.98902777777778,
            'time_in_flow_class': <Swig Object of type 'DateTime *' at 0x7f7d591269a0>,
            'time_inlet_control': 0.0,
            'time_normal_flow': 0.0001388888888888889,
            'time_surcharged': 0.0
        }

        """
        return self._model.conduit_statistics(self.linkid)


class Pump(Link):
    """
    Pump Object: Subclass of Link Object.
    """

    def __init__(self):
        super(Pump, self).__init__()

    @property
    def pump_statistics(self):
        """
        Pump Stats. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +------------------+
        | percent_utilized |
        +------------------+
        | min_flowrate     |
        +------------------+
        | average_flowrate |
        +------------------+
        | max_flowrate     |
        +------------------+
        | total_volume     |
        +------------------+
        | energy_consumed  |
        +------------------+
        | off_curve_low    |
        +------------------+
        | off_curve_high   |
        +------------------+
        | number_startups  |
        +------------------+
        | total_periods    |
        +------------------+

        :return: Group of Stats
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Links
        >>> import pprint
        >>> pp = pprint.PrettyPrinter(indent=4)
        >>>
        >>> with Simulation('tests/data/model_pump_setting.inp') as sim:
        ...     c3 = Links(sim)["C3"]
        ...     for step in sim:
        ...         pass
        ...     pp.pprint(c3.pump_statistics)
        o  Retrieving project data
        {   'average_flowrate': 20.0,
            'energy_consumed': 2792.5975765384896,
            'max_flowrate': 20.0,
            'min_flowrate': 0.0,
            'number_startups': 1,
            'off_curve_high': 0.0,
            'off_curve_low': 0.0,
            'percent_utilized': 208800.0,
            'total_periods': 208801,
            'total_volume': 4176000.0
        }
        """
        return self._model.pump_statistics(self.linkid)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2018 Jennifer Wu
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from pyswmm.swmm5 import PYSWMMException
from pyswmm import LidControls
from pyswmm.toolkitapi import ObjectType, LidUParams, LidUOptions, LidResults
from pyswmm.lidunits import Surface, Pavement, Soil, Storage, WaterBalance


class LidGroups(object):
    """
    LidGroups Iterator Methods.

    :param object model: Open Model Instance

    This allows the user to iterate and get one of the LID Groups as
    defined inside SWMM's [LID_USAGE] section.  For example, here is a sample
    section from a SWMM INP file.

    .. code-block::

        [LID_USAGE]
        ;;Subcatchment   LID Process   Number  Area    Width   InitSat  FromImp  ToPerv   RptFile  DrainTo  FromPerv
        ;;-------------- ------------- ------- ------- ------- -------- -------- -------- -------- -------- ----------
        J1_J3_qqqqqqq    LID_C1        1       180.00  2       0        0        0        *        *        0
        J11_J14_qqqqqqq  LID_C1        1       140     2       0        0        0        *        *        0


    Now we can peform manipulations in PySWMM and get objects.

    .. code-block:: python

        from pyswmm import Simulation, LidGroups

        with Simulation('lid_model.inp') as sim:
            for lid_group in LidGroups(sim):
                print(lid_group)

    .. code-block::

        >>> J1_J3_qqqqqqq
        >>> J11_J14_qqqqqqq

    Once a handle has been created of an LidGroup object we can now make
    adjustments to the parameters and change the behavior.  The following example
    provides some context around the different object handles created.  One
    feature presented herein shows that during a simulation the drain_node
    (or drain subcatchement) can be changed during the simulation.

    .. code-block:: python

        from pyswmm import Simulation, LidGroups

        with Simulation('lid_model.inp') as sim:
            # This returns an LidGroup object
            lid_group_j1_j3 = LidGroups(sim)["J1_J3_qqqqqqq"]

            # Next to get access to one of the LID units
            # inside this subcatchment it is indexed.

            # This returns an LidUnit object
            lid_unit0=lid_group_j1_j3[0]

            for ind, step in enumerate(sim):
                if ind == 50:
                    lid_unit0.drain_node = "J04"

    Continuing on the previous example, either during the simulation or after
    it completes, the user has access to lid unit results.  The following code
    provides details on how to access differne LidUnit results

    .. code-block:: python

        from pyswmm import Simulation, LidGroups

        with Simulation('lid_model.inp') as sim:
            # This returns an LidGroup object
            lid_group_j1_j3 = LidGroups(sim)["J1_J3_qqqqqqq"]

            # Next to get access to one of the LID units
            # inside this subcatchment it is indexed.

            # This returns an LidUnit object
            lid_unit0=lid_group_j1_j3[0]

            for ind, step in enumerate(sim):
                if ind == 50:
                    lid_unit0.drain_node = "J04"

            surface = lid_unit0.surface
            pavement = lid_unit0.pavement
            soil = lid_unit0.soil
            storage = lid_unit0.storage
            water_balance = lid_unit0.water_balance
            print(water_balance.inflow)
            print(water_balance.evaporation)
            print(water_balance.infiltration)
            print(water_balance.surface_flow)
            print(water_balance.drain_flow)
            print(water_balance.initial_volume)
            print(water_balance.final_volume)
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model
        self._cuindex = 0
        self._nLidGroups = self._model.getProjectSize(ObjectType.SUBCATCH.value)

    def __len__(self):
        """
        Return number of defined LidGroups.
        Note that there are one lidgroup per subcatchment
        Use the expression 'len(LidGroups)'.

        :return: Number of LidGroups
        :rtype: int

        """
        return self._model.getProjectSize(ObjectType.SUBCATCH.value)

    def __contains__(self, subcatchmentid):
        """
        Checks if Subcatchment ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.SUBCATCH.value, subcatchmentid)

    def __getitem__(self, subcatchmentid):
        if self.__contains__(subcatchmentid):
            return LidGroup(self._model, subcatchmentid)
        else:
            raise PYSWMMException("Subcatchment ID Does not Exist")

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nLidGroups:
            lidgroupobject = self.__getitem__(self._subcatchmentid)
            self._cuindex += 1  # Next Iteration
            return lidgroupobject
        else:
            raise StopIteration()

    @property
    def _subcatchmentid(self):
        """Subcatchment ID."""
        return self._model.getObjectId(ObjectType.SUBCATCH.value, self._cuindex)


class LidGroup(object):
    """
    Methods and properties associated with a group of LIDs defined on a subcatchment

    :param object model: Open Model Instance
    :param str subcatchmentid: Name of subcatchment associated with the Lid Group
    """

    def __init__(self, model, subcatchmentid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if subcatchmentid not in model.getObjectIDList(ObjectType.SUBCATCH.value):
            raise PYSWMMException("Subcatchment ID Does not Exist")
        self._model = model
        self._subcatchmentid = subcatchmentid
        self._cuindex = 0
        self._nLidUnits = model.getLidUCount(subcatchmentid)

    def __str__(self):
        return self._subcatchmentid

    def __len__(self):
        """
        Return number of defined LidUnit per LidGroup.
        Note that there can be multiple LidUnit per subcatchment
        Use the expression 'len(LidGroup)'.

        :return: Number of LidUnit
        :rtype: int

        """
        return self._nLidUnits

    def __contains__(self, index):
        """
        Checks if Subcatchment ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return index < self._nLidUnits

    def __getitem__(self, index):
        if self.__contains__(index):
            return LidUnit(self._model, self._subcatchmentid, index)
        else:
            raise PYSWMMException("Lid Unit Does Not Exist")

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nLidUnits:
            lidunitobject = self.__getitem__(self._cuindex)
            self._cuindex += 1  # Next Iteration
            return lidunitobject
        else:
            raise StopIteration()

    @property
    def pervious_area(self):
        """
        Get lid group amount of pervious area

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidGResult(
            self._subcatchmentid, LidResults.pervArea.value
        )

    @property
    def flow_to_pervious(self):
        """
        Get lid group total flow sent to pervious area

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidGResult(
            self._subcatchmentid, LidResults.flowToPerv.value
        )

    @property
    def old_drain_flow(self):
        """
        Get lid group total drain flow in pervious period

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidGResult(
            self._subcatchmentid, LidResults.oldDrainFlow.value
        )

    @property
    def new_drain_flow(self):
        """
        Get lid group total drain flow in current period

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidGResult(
            self._subcatchmentid, LidResults.newDrainFlow.value
        )


class LidUnit(object):
    """
    Lid Unit Methods.

    :param object model: Open Model Instance
    :param str subcatchmentid: Subcatchment ID
    :param str lidid: Lid unit ID

    Lid Unit Parameters

    +--------------------+--------------------+--------------------+--------------------+
    | Parameter          | Getter             | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | subcatchment       | enabled            | disabled           | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | lid_control        | enabled            | disabled           | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | unit_area          | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | full_width         | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | initial_saturation | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | from_impervious    | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | from_pervious      | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | index              | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | number             | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | to_pervious        | enabled            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | drain_subcatchment | enabled            | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | drain_node         | enabled            | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    """

    def __init__(self, model, subcatchmentid, lidid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model
        self._subcatchmentid = subcatchmentid
        self._lidid = lidid

        self.surface = Surface(model, self)
        self.pavement = Pavement(model, self)
        self.soil = Soil(model, self)
        self.storage = Storage(model, self)
        self.water_balance = WaterBalance(model, self)

    @property
    def subcatchment(self):
        return self._subcatchmentid

    @property
    def lid_control(self):
        index = self.index
        return self._model.getObjectId(ObjectType.LID.value, index)

    @property
    def unit_area(self):
        """
        Get lid unit area.

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.unitArea.value
        )

    @unit_area.setter
    def unit_area(self, param):
        """Set lid unit area"""
        return self._model.setLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.unitArea.value, param
        )

    @property
    def full_width(self):
        """
        Get lid unit full top width.

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fullWidth.value
        )

    @full_width.setter
    def full_width(self, param):
        """Set lid unit full top width."""
        return self._model.setLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fullWidth.value, param
        )

    @property
    def initial_saturation(self):
        """
        Get lid initial saturation of soil and storage layers.

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.initSat.value
        )

    @initial_saturation.setter
    def initial_saturation(self, param):
        """Set lid initial saturation of soil and storage layers."""
        return self._model.setLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.initSat.value, param
        )

    @property
    def from_impervious(self):
        """
        Get lid fraction of impervious area runoff treated

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fromImperv.value
        )

    @from_impervious.setter
    def from_impervious(self, param):
        """Set lid fraction of impervious area runoff treated"""
        return self._model.setLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fromImperv.value, param
        )

    @property
    def from_pervious(self):
        """
        Get lid fraction of pervious area runoff treated

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fromPerv.value
        )

    @from_pervious.setter
    def from_pervious(self, param):
        """Set lid fraction of pervious area runoff treated"""
        return self._model.setLidUParam(
            self._subcatchmentid, self._lidid, LidUParams.fromPerv.value, param
        )

    @property
    def index(self):
        """
        Get lid control index

        :return: Parameter Value
        :rtype: int
        """
        return self._model.getLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.index.value
        )

    @index.setter
    def index(self, param):
        """Set lid control index"""
        if isinstance(param, str) and self._model.ObjectIDexist(
            ObjectType.LID.value, param
        ):
            controlIndex = self._model.getObjectIDIndex(ObjectType.LID.value, param)
        elif (
            isinstance(param, int)
            and param >= 0
            and param < self._model.getProjectSize(ObjectType.LID.value)
        ):
            controlIndex = param
        else:
            controlIndex = 0
            raise PYSWMMException("Invalid Input")

        return self._model.setLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.index.value, controlIndex
        )

    @property
    def number(self):
        """
        Get lid number of replicate units

        :return: Parameter Value
        :rtype: int
        """
        return self._model.getLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.number.value
        )

    @number.setter
    def number(self, param):
        """Set lid number of replicate units"""
        return self._model.setLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.number.value, param
        )

    @property
    def to_pervious(self):
        """
        Get lid to pervious area (1 if outflow sent to pervious area)
                                 (0 if not)

        :return: Parameter Value
        :rtype: int
        """
        return self._model.getLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.toPerv.value
        )

    @to_pervious.setter
    def to_pervious(self, param):
        """
        Set lid to pervious area (1 if outflow sent to pervious area)
                                 (0 if not)
        """
        return self._model.setLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.toPerv.value, param
        )

    @property
    def drain_subcatchment(self):
        """
        Get lid drain to subcatchment index
        Negative if no recieving subcatchment

        :return: Parameter Value
        :rtype: int
        """
        return self._model.getLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.drainSub.value
        )

    @drain_subcatchment.setter
    def drain_subcatchment(self, param):
        """Set lid drain to subcatchment index"""
        if isinstance(param, str) and self._model.ObjectIDexist(
            ObjectType.SUBCATCH.value, param
        ):
            subIndex = self._model.getObjectIDIndex(ObjectType.SUBCATCH.value, param)
        elif (
            isinstance(param, int)
            and param >= -1
            and param < self._model.getProjectSize(ObjectType.SUBCATCH.value)
        ):
            subIndex = param
        else:
            subIndex = 0
            raise PYSWMMException("Invalid Input")

        self._model.setLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.drainSub.value, subIndex
        )

    @property
    def drain_node(self):
        """
        Get lid drain to node index
        Negative if no recieving node

        :return: Parameter Value
        :rtype: int
        """
        return self._model.getLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.drainNode.value
        )

    @drain_node.setter
    def drain_node(self, param):
        """Set lid drain to node index"""
        if isinstance(param, str) and self._model.ObjectIDexist(
            ObjectType.NODE.value, param
        ):
            nodeIndex = self._model.getObjectIDIndex(ObjectType.NODE.value, param)
        elif (
            isinstance(param, int)
            and param >= -1
            and param < self._model.getProjectSize(ObjectType.NODE.value)
        ):
            nodeIndex = param
        else:
            nodeIndex = 0
            raise PYSWMMException("Invalid Input")

        self._model.setLidUOption(
            self._subcatchmentid, self._lidid, LidUOptions.drainNode.value, nodeIndex
        )

    @property
    def dry_time(self):
        """
        Get lid time since last rainfall (sec)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(
            self._subcatchmentid, self._lidid, LidResults.dryTime.value
        )

    @property
    def old_drain_flow(self):
        """
        Get lid pervious drain flow

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(
            self._subcatchmentid, self._lidid, LidResults.oldDrainFlow.value
        )

    @property
    def new_drain_flow(self):
        """
        Get lid current drain flow

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(
            self._subcatchmentid, self._lidid, LidResults.newDrainFlow.value
        )

    @property
    def evaporation(self):
        """
        Get lid current evaporation rate

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(
            self._subcatchmentid, self._lidid, LidResults.evapRate.value
        )

    @property
    def native_infiltration(self):
        """
        Get lid native infilration rate limit

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(
            self._subcatchmentid, self._lidid, LidResults.nativeInfil.value
        )



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Subcatchments module for the pythonic interface to SWMM5."""

# Local imports
from pyswmm.swmm5 import PYSWMMException
from pyswmm.toolkitapi import ObjectType, SubcParams, SubcPollut, SubcResults


class Subcatchments(object):
    """
    Subcatchment Iterator Methods.

    :param object model: Open Model Instance

    Examples:

    >>> from pyswmm import Simulation, Subcatchments
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     for subcatchment in Subcatchments(sim):
    ...         print(subcatchment)
    ...         print(subcatchment.subcatchmentid)
    <swmm5.Subcatchment object at 0x031B0350>
    S1
    <swmm5.Subcatchment object at 0x030693D0>
    S2
    <swmm5.Subcatchment object at 0x031B0350>
    S3
    <swmm5.Subcatchment object at 0x030693D0>
    S4

    Iterating over Subcatchments Object

    >>> subcatchments = Subcatchments(sim)
    >>> for subcatchment in subcatchments:
    ...     print(subcatchment.subcatchmentid)
    S0
    S1
    S2
    S3

    Testing Existence

    >>> subcatchments = Subcatchments(sim)
    >>> "S1" in subcatchments
    True

    Initializing a subcatchment Object

    >>> subcatchments = Subcatchments(sim)
    >>> s1 = subcatchments['S1']
    >>> print(s1.area)
    12
    >>>
    >>> s1.area = 200
    >>> print(s1.area)
    200
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model
        self._cuindex = 0
        self._nSubcatchments = self._model.getProjectSize(ObjectType.SUBCATCH.value)

    def __len__(self):
        """
        Return number of subcatchments.

        Use the expression 'len(Subcatchments)'.

        :return: Number of Subcatchments
        :rtype: int

        """
        return self._model.getProjectSize(ObjectType.SUBCATCH.value)

    def __contains__(self, subcatchmentid):
        """
        Checks if Subcatchment ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.SUBCATCH.value, subcatchmentid)

    def __getitem__(self, subcatchmentid):
        if self.__contains__(subcatchmentid):
            return Subcatchment(self._model, subcatchmentid)
        else:
            raise PYSWMMException("Subcatchment ID Does not Exist")

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nSubcatchments:
            subcatchmentobject = self.__getitem__(self._subcatchmentid)
            self._cuindex += 1  # Next Iteration
            return subcatchmentobject
        else:
            raise StopIteration()

    @property
    def _subcatchmentid(self):
        """Subcatchment ID."""
        return self._model.getObjectId(ObjectType.SUBCATCH.value, self._cuindex)


class Subcatchment(object):
    """
    Subcatchment Methods.

    :param object model: Open Model Instance
    :param str subcatchmentid: Subcatchment ID

    Examples:

    >>> from pyswm import Simulation, Subcatchments
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     s1 = Subcatchments(sim)["S1"]
    ...     print(s1.rainfall)
    ...     for step in sim:
    ...         print(s1.rainfall)
    0.04
    """

    def __init__(self, model, subcatchmentid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if subcatchmentid not in model.getObjectIDList(ObjectType.SUBCATCH.value):
            raise PYSWMMException("ID Not valid")
        self._model = model
        self._subcatchmentid = subcatchmentid

    # --- Get Parameters
    # -------------------------------------------------------------------------
    @property
    def subcatchmentid(self):
        """
        Get Subcatchment ID.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.subcatchmentid)
        S1
        """
        return self._subcatchmentid

    @property
    def connection(self):
        """
        Get Subcatchment Outlet Connection.

        This function return the type of loading surface and the ID. The two
        load to objects are nodes and other subcatchments.

        +--------------+---+
        | Node         | 2 |
        +--------------+---+
        | Subcatchment | 1 |
        +--------------+---+

        :return: (Loading Surface Type, ID)
        :rtype: tuple

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.connection)
        (2, 'J2')
        """
        return self._model.getSubcatchOutConnection(self._subcatchmentid)

    @property
    def width(self):
        """
        Get/set subcatchment width.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.width)
        100.0

        Setting the value

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.width)
        ...     s1.width = 30
        ...     print(s1.width)
        100
        30
        """
        return self._model.getSubcatchParam(
            self._subcatchmentid, SubcParams.width.value
        )

    @width.setter
    def width(self, param):
        """Set Subcatchment Width."""
        self._model.setSubcatchParam(
            self._subcatchmentid, SubcParams.width.value, param
        )

    @property
    def area(self):
        """
        Get/set subcatchment area.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.area)
        10

        Setting the value

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.area)
        ...     s1.area = 50
        ...     print(s1.area)
        10
        50
        """
        return self._model.getSubcatchParam(self._subcatchmentid, SubcParams.area.value)

    @area.setter
    def area(self, param):
        """Set Subcatchment area."""
        self._model.setSubcatchParam(self._subcatchmentid, SubcParams.area.value, param)

    @property
    def percent_impervious(self):
        """
        Get/set subcatchment percent impervious.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.percent_impervious)
        10

        Setting the value

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.percent_impervious)
        ...     s1.percent_impervious = 50
        ...     print(s1.percent_impervious)
        10
        50
        """
        return self._model.getSubcatchParam(
            self._subcatchmentid, SubcParams.fracImperv.value
        )

    @percent_impervious.setter
    def percent_impervious(self, param):
        """Set Subcatchment percent impervious."""
        self._model.setSubcatchParam(
            self._subcatchmentid, SubcParams.fracImperv.value, param
        )

    @property
    def slope(self):
        """
        Get/set subcatchment slope.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.slope)
        0.01

        Setting the value

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.slope)
        ...     s1.slope = 0.02
        ...     print(s1.slope)
        0.1
        0.2
        """
        return self._model.getSubcatchParam(
            self._subcatchmentid, SubcParams.slope.value
        )

    @slope.setter
    def slope(self, param):
        """Set Subcatchment Ponding Area."""
        self._model.setSubcatchParam(
            self._subcatchmentid, SubcParams.slope.value, param
        )

    @property
    def curb_length(self):
        """
        Get/set subcatchment curb length.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.curb_length)
        0

        Setting the value

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     print(s1.curb_length)
        ...     s1.curb_length = 100
        ...     print(s1.curb_length)
        0
        100
        """
        return self._model.getSubcatchParam(
            self._subcatchmentid, SubcParams.curbLength.value
        )

    @curb_length.setter
    def curb_length(self, param):
        """Set Subcatchment curb length."""
        self._model.setSubcatchParam(
            self._subcatchmentid, SubcParams.curbLength.value, param
        )

    @property
    def rainfall(self):
        """
        Get Subcatchment Results for Rainfall.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.rainfall)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.rainfall.value
        )

    @property
    def evaporation_loss(self):
        """
        Get Subcatchment Results for evaporation loss.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.evaporation_loss)
        0.01
        0.01
        0.01
        0.01
        0.01
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.evapLoss.value
        )

    @property
    def infiltration_loss(self):
        """
        Get Subcatchment Results for Infiltration Loss.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.infiltration_loss)
        0
        0.01
        0.01
        0.01
        0.01
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.infilLoss.value
        )

    @property
    def runon(self):
        """
        Get Subcatchment Results for Run On.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.runon)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.runon.value
        )

    @property
    def runoff(self):
        """
        Get Subcatchment Results for Run Off Rate.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.runoff)
        0
        0
        0.01
        0
        0
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.newRunoff.value
        )

    @property
    def snow_depth(self):
        """
        Get Subcatchment Results for Snow Depth.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.snow_depth)
        0
        0.5
        0.51
        0.52
        0.49
        """
        return self._model.getSubcatchResult(
            self._subcatchmentid, SubcResults.newSnowDepth.value
        )

    @property
    def buildup(self):
        """
        Get Pollutant Results for Surface Buildup on a Subcatchment.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Subcatchment Surface Buildup Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/buildup-test.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.buildup)
        {'test-pollutant': 8.0}
        {'test-pollutant': 8.0}
        {'test-pollutant': 7.998}
        {'test-pollutant': 7.991}
        {'test-pollutant': 7.981}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        buildup_array = self._model.getSubcatchPollut(
            self._subcatchmentid, SubcPollut.buildup.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = buildup_array[ind]

        return out_dict

    @property
    def conc_ponded(self):
        """
        Get Pollutant Results for Current Concentration of Pollutant in Ponded
        Water on a Subcatchment.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Subcatchment Ponded Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/buildup-test.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.conc_ponded)
        {'test-pollut1': 0.0, 'test-pollut2': 0.0}
        {'test-pollut1': 0.0, 'test-pollut2': 0.0}
        {'test-pollut1': 0.0, 'test-pollut2': 0.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        cPonded_array = self._model.getSubcatchPollut(
            self._subcatchmentid, SubcPollut.concPonded.value
        )
        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = cPonded_array[ind]

        return out_dict

    @property
    def pollut_quality(self):
        """
        Get Current Pollutant Water Quality Results from Subcatchment Runoff.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Subcatchment Runoff Pollutant Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/buildup-test.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.pollut_quality)
        {'TSS': 0.0, 'Lead': 0.0}
        {'TSS': 0.0, 'Lead': 0.0}
        {'TSS': 0.0, 'Lead': 0.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        subcqual_array = self._model.getSubcatchPollut(
            self._subcatchmentid, SubcPollut.subcQual.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = subcqual_array[ind]

        return out_dict

    @property
    def runoff_total_loading(self):
        """
        Get Total Pollutant Loading from Subcatchment Runoff.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Subcatchment Runoff Pollutant Loading Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Subcatchments
        >>>
        >>> with Simulation('tests/buildup-test.inp') as sim:
        ...     s1 = Subcatchments(sim)["S1"]
        ...     for step in sim:
        ...         print(s1.runoff_total_loading)
        {'TSS': 0.01, 'Lead': 0.001}
        {'TSS': 0.02, 'Lead': 0.002}
        {'TSS': 0.03, 'Lead': 0.003}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        totalload_array = self._model.getSubcatchPollut(
            self._subcatchmentid, SubcPollut.subcTotalLoad.value
        )

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = totalload_array[ind]

        return out_dict

    @property
    def statistics(self):
        """
        Subcatchment Flow Stats. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +-------------------+
        | precipitation     |
        +-------------------+
        | runon             |
        +-------------------+
        | evaporation       |
        +-------------------+
        | infiltration      |
        +-------------------+
        | runoff            |
        +-------------------+
        | peak_runoff_rate  |
        +-------------------+

        :return: Group of Stats
        :rtype: dict
        """
        return self._model.subcatch_statistics(self.subcatchmentid)



class SimulationContextWarning(ResourceWarning):
    """Warning raised for SWMM Simulation Context

    Attributes:
        message -- explanation of the error
    """

    message = """
    \tThe Simulation object is intended to be used with a context
    \tmanager. System resources will not be freed without calling 
    \tthe close method. See Simulation class docs for details.
    """



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Pytest fixtures and helpers."""


def get_model_files(name):
    """Return the input, report and output name based on the input name."""
    base_name = name.split(".inp")[0]
    report_name = base_name + ".rpt"
    output_name = base_name + ".out"
    return name, report_name, output_name



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""SWMM Object Enum."""
from swmm.toolkit import shared_enum

# Standard library imports
from enum import Enum
import ctypes


class SimulationTime(Enum):
    StartDateTime = 0
    EndDateTime = 1
    ReportStart = 2


class SimulationUnits(Enum):
    UnitSystem = 0  # (0-US, 1-SI)
    FlowUnits = 1


class SimAnalysisSettings(Enum):
    AllowPonding = 0  # No ponding at nodes
    SkipSteadyState = 1  # Do flow routing in steady state periods
    IgnoreRainfall = 2  # Analyze rainfall/runoff
    IgnoreRDII = 3  # Analyze RDII
    IgnoreSnowmelt = 4  # Analyze snowmelt
    IgnoreGwater = 5  # Analyze groundwater
    IgnoreRouting = 6  # Analyze flow routing
    IgnoreQuality = 7  # Analyze water quality


class SimulationParameters(Enum):
    RouteStep = 0  # Routing time step (sec)
    MinRouteStep = 1  # Minimum variable time step (sec)
    LengtheningStep = 2  # Time step for lengthening (sec)
    StartDryDays = 3  # Antecedent dry days
    CourantFactor = 4  # Courant time step factor
    MinSurfArea = 5  # Minimum nodal surface area
    MinSlope = 6  # Minimum conduit slope
    RunoffError = 7  # Runoff continuity error
    GwaterError = 8  # Groundwater continuity error
    FlowError = 9  # Flow routing error
    QualError = 10  # Quality routing error
    HeadTol = 11  # DW routing head tolerance (ft)
    SysFlowTol = 12  # Tolerance for steady system flow
    LatFlowTol = 13  # Tolerance for steady nodal inflow


class ObjectType(Enum):
    GAGE = 0  # rain gage
    SUBCATCH = 1  # subcatchment
    NODE = 2  # conveyance system node
    LINK = 3  # conveyance system link
    POLLUT = 4  # pollutant
    LANDUSE = 5  # land use category
    TIMEPATTERN = 6  # dry weather flow time pattern
    CURVE = 7  # generic table of values
    TSERIES = 8  # generic time series of values
    CONTROL = 9  # conveyance system control rules
    TRANSECT = 10  # irregular channel cross-section
    AQUIFER = 11  # groundwater aquifer
    UNITHYD = 12  # RDII unit hydrograph
    SNOWMELT = 13  # snowmelt parameter set
    SHAPE = 14  # custom conduit shape
    LID = 15  # LID treatment units
    MAX_OBJ_TYPES = 16  # MaximumObjectTypes


class NodeParams(Enum):
    invertElev = 0  # double
    fullDepth = 1  # double
    surDepth = 2  # double
    pondedArea = 3  # double
    initDepth = 4  # double


class NodeResults(Enum):
    totalinflow = 0  # Total Inflow
    outflow = 1  # Total Outflow
    losses = 2  # Losses (evap + exfiltration loss)
    newVolume = 3  # Current Volume
    overflow = 4  # overflow
    newDepth = 5  # Current water depth
    newHead = 6  # Current water head
    newLatFlow = 7  # Current Lateral Inflow
    hyd_res_time = 8  # Current hydraulic retention time


class NodePollut(Enum):
    nodeQual = 0  # Current Water Quality Value
    inflowQual = 1  # Current inflow water quality value
    reactorQual = 2  # Current quality value in the mixed reactor


class NodeType(Enum):
    junction = 0  # Junction Type
    outfall = 1  # Outfall Type
    storage = 2  # Storage Type
    divider = 3  # Divider Type


class LinkParams(Enum):
    offset1 = 0  # double
    offset2 = 1  # double
    q0 = 2  # double
    qLimit = 3  # double
    cLossInlet = 4  # double
    cLossOutlet = 5  # double
    cLossAvg = 6  # double
    seepRate = 7  # double


class LinkResults(Enum):
    newFlow = 0  # double
    newDepth = 1  # double
    newVolume = 2  # double
    surfArea1 = 3  # double
    surfArea2 = 4  # double
    setting = 5  # double
    targetSetting = 6  # double
    froude = 7  # double


class LinkPollut(Enum):
    linkQual = 0  # Current Water Quality Value
    totalLoad = 1  # Total Quality Mass Loading
    reactorQual = 2  # Current quality value in the mixed reactor


class LinkType(Enum):
    conduit = 0  # Conduit Type
    pump = 1  # Pump Type
    orifice = 2  # Orifice Type
    weir = 3  # Weir Type
    outlet = 4  # Outlet Type


class SubcParams(Enum):
    width = 0  # double
    area = 1  # double
    fracImperv = 2  # double
    slope = 3  # double
    curbLength = 4  # double
    # initBuildup = 5  # double


class SubcResults(Enum):
    rainfall = 0  # Current Rainfall
    evapLoss = 1  # Current Evaporation Loss
    infilLoss = 2  # Current Infiltration Loss
    runon = 3  # Subcatchment Runon
    newRunoff = 4  # Current Runoff
    newSnowDepth = 5  # Current Snow Depth


class SubcPollut(Enum):
    buildup = 0  # Subcatchment Surface Buildup
    concPonded = 1  # Ponded Pollutant Concentration
    subcQual = 2  # Current Pollutant Runoff Quality
    subcTotalLoad = 3  # Total Pollutant Washoff Load


class LidUParams(Enum):
    unitArea = 0  # double
    fullWidth = 1  # double
    botWidth = 2  # double
    initSat = 3  # double
    fromImperv = 4  # double
    fromPerv = 5  # double


class LidResults(Enum):
    inflow = 0  # double
    evap = 1  # double
    infil = 2  # double
    surfFlow = 3  # double
    drainFlow = 4  # double
    initVol = 5  # double
    finalVol = 6  # double
    surfDepth = 7  # double
    paveDepth = 8  # double
    soilMoist = 9  # double
    storDepth = 10  # double
    dryTime = 11  # double
    oldDrainFlow = 12  # double
    newDrainFlow = 13  # double
    pervArea = 14  # double
    flowToPerv = 15  # double
    evapRate = 16  # double
    nativeInfil = 17  # double
    surfInflow = 18  # double
    surfInfil = 19  # double
    surfEvap = 20  # double
    surfOutflow = 21  # double
    paveEvap = 22  # double
    pavePerc = 23  # double
    soilEvap = 24  # double
    soilPerc = 25  # double
    storInflow = 26  # double
    storExfil = 27  # double
    storEvap = 28  # double
    storDrain = 29  # double


class LidUOptions(Enum):
    index = 0  # int
    number = 1  # int
    toPerv = 2  # int
    drainSub = 3  # int
    drainNode = 4  # int


class LidLayers(Enum):
    surface = 0  # int
    soil = 1  # int
    storage = 2  # int
    pavement = 3  # int
    drain = 4  # int
    drainMat = 5  # int


class LidLayersProperty(Enum):
    thickness = 0  # double
    voidFrac = 1  # double
    roughness = 2  # double
    surfSlope = 3  # double
    sideSlope = 4  # double
    alpha = 5  # double
    porosity = 6  # double
    fieldCap = 7  # double
    wiltPoint = 8  # double
    suction = 9  # double
    kSat = 10  # double
    kSlope = 11  # double
    clogFactor = 12  # double
    impervFrac = 13  # double
    coeff = 14  # double
    expon = 15  # double
    offset = 16  # double
    delay = 17  # double
    hOpen = 18  # double
    hClose = 19  # double
    qCurve = 20  # integer
    regenDays = 21  # double
    regenDegree = 22  # double


class RainGageResults(Enum):
    total_precip = 0
    rainfall = 1
    snowfall = 2


class HotstartFile(Enum):
    use = 0
    save = 1


class NodeStats(ctypes.Structure):
    _fields_ = [
        ("avgDepth", ctypes.c_double),
        ("maxDepth", ctypes.c_double),
        ("maxDepthDate", ctypes.c_double),
        ("maxRptDepth", ctypes.c_double),
        ("volFlooded", ctypes.c_double),
        ("timeFlooded", ctypes.c_double),
        ("timeSurcharged", ctypes.c_double),
        ("timeCourantCritical", ctypes.c_double),
        ("totLatFlow", ctypes.c_double),
        ("maxLatFlow", ctypes.c_double),
        ("maxInflow", ctypes.c_double),
        ("maxOverflow", ctypes.c_double),
        ("maxPondedVol", ctypes.c_double),
        ("maxInflowDate", ctypes.c_double),
        ("maxOverflowDate", ctypes.c_double),
    ]
    _py_alias_ids = {
        "avgDepth": "average_depth",
        "maxDepth": "max_depth",
        "maxDepthDate": "max_depth_date",
        "maxRptDepth": "max_report_depth",
        "volFlooded": "flooding_volume",
        "timeFlooded": "flooding_duration",
        "timeSurcharged": "surcharge_duration",
        "timeCourantCritical": "courant_crit_duration",
        "totLatFlow": "lateral_infow_vol",
        "maxLatFlow": "peak_lateral_inflowrate",
        "maxInflow": "peak_total_inflow",
        "maxOverflow": "peak_flooding_rate",
        "maxPondedVol": "max_ponded_volume",
        "maxInflowDate": "max_inflow_date",
        "maxOverflowDate": "max_flooding_date",
        "loads": "pollutant_loading",
    }


class StorageStats(ctypes.Structure):
    _fields_ = [
        ("initVol", ctypes.c_double),
        ("avgVol", ctypes.c_double),
        ("maxVol", ctypes.c_double),
        ("maxFlow", ctypes.c_double),
        ("evapLosses", ctypes.c_double),
        ("exfilLosses", ctypes.c_double),
        ("maxVolDate", ctypes.c_double),
    ]
    _py_alias_ids = {
        "initVol": "initial_volume",
        "avgVol": "average_volume",
        "maxVol": "max_volume",
        "maxFlow": "peak_flowrate",
        "evapLosses": "evap_loss",
        "exfilLosses": "exfil_loss",
        "maxVolDate": "max_vol_date",
    }


PollutArray = ctypes.POINTER(ctypes.c_double)


class OutfallStats(ctypes.Structure):
    _fields_ = [
        ("avgFlow", ctypes.c_double),
        ("maxFlow", ctypes.c_double),
        ("totalLoad", PollutArray),
        ("totalPeriods", ctypes.c_int),
    ]
    _py_alias_ids = {
        "avgFlow": "average_flowrate",
        "maxFlow": "peak_flowrate",
        "totalLoad": "pollutant_loading",
        "totalPeriods": "total_periods",
        "loads": "pollutant_loading",
    }


FlowClassArray = ctypes.c_double * 7


class LinkStats(ctypes.Structure):
    _fields_ = [
        ("maxFlow", ctypes.c_double),
        ("maxFlowDate", ctypes.c_double),
        ("maxVeloc", ctypes.c_double),
        ("maxDepth", ctypes.c_double),
        ("timeNormalFlow", ctypes.c_double),
        ("timeInletControl", ctypes.c_double),
        ("timeSurcharged", ctypes.c_double),
        ("timeFullUpstream", ctypes.c_double),
        ("timeFullDnstream", ctypes.c_double),
        ("timeFullFlow", ctypes.c_double),
        ("timeCapacityLimited", ctypes.c_double),
        ("timeInFlowClass", FlowClassArray),
        ("timeCourantCritical", ctypes.c_double),
        ("flowTurns", ctypes.c_long),
        ("flowTurnSign", ctypes.c_int),
    ]
    _py_alias_ids = {
        "maxFlow": "peak_flow",
        "maxFlowDate": "peak_flow_date",
        "maxVeloc": "peak_velocity",
        "maxDepth": "peak_depth",
        "timeNormalFlow": "time_normal_flow",
        "timeInletControl": "time_inlet_control",
        "timeSurcharged": "time_surcharged",
        "timeFullUpstream": "time_full_upstream",
        "timeFullDnstream": "time_full_downstream",
        "timeFullFlow": "time_full_flow",
        "timeCapacityLimited": "time_capacity_limited",
        "timeInFlowClass": "time_in_flow_class",
        "timeCourantCritical": "time_courant_crit",
        "flowTurns": "flow_turns",
        "flowTurnSign": "flow_turn_sign",
    }


class PumpStats(ctypes.Structure):
    _fields_ = [
        ("utilized", ctypes.c_double),
        ("minFlow", ctypes.c_double),
        ("avgFlow", ctypes.c_double),
        ("maxFlow", ctypes.c_double),
        ("volume", ctypes.c_double),
        ("energy", ctypes.c_double),
        ("offCurveLow", ctypes.c_double),
        ("offCurveHigh", ctypes.c_double),
        ("startUps", ctypes.c_int),
        ("totalPeriods", ctypes.c_int),
    ]
    _py_alias_ids = {
        "utilized": "percent_utilized",
        "minFlow": "min_flowrate",
        "avgFlow": "average_flowrate",
        "maxFlow": "max_flowrate",
        "volume": "total_volume",
        "energy": "energy_consumed",
        "offCurveLow": "off_curve_low",
        "offCurveHigh": "off_curve_high",
        "startUps": "number_startups",
        "totalPeriods": "total_periods",
    }


class SubcStats(ctypes.Structure):
    _fields_ = [
        ("precip", ctypes.c_double),
        ("runon", ctypes.c_double),
        ("evap", ctypes.c_double),
        ("infil", ctypes.c_double),
        ("runoff", ctypes.c_double),
        ("maxFlow", ctypes.c_double),
    ]
    _py_alias_ids = {
        "precip": "precipitation",
        "runon": "runon",
        "evap": "evaporation",
        "infil": "infiltration",
        "runoff": "runoff",
        "maxFlow": "peak_runoff_rate",
    }


class RoutingTotals(ctypes.Structure):
    _fields_ = [
        ("dwInflow", ctypes.c_double),
        ("wwInflow", ctypes.c_double),
        ("gwInflow", ctypes.c_double),
        ("iiInflow", ctypes.c_double),
        ("exInflow", ctypes.c_double),
        ("flooding", ctypes.c_double),
        ("outflow", ctypes.c_double),
        ("evapLoss", ctypes.c_double),
        ("seepLoss", ctypes.c_double),
        ("reacted", ctypes.c_double),
        ("initStorage", ctypes.c_double),
        ("finalStorage", ctypes.c_double),
        ("pctError", ctypes.c_double),
    ]
    _py_alias_ids = {
        "dwInflow": "dry_weather_inflow",
        "wwInflow": "wet_weather_inflow",
        "gwInflow": "groundwater_inflow",
        "iiInflow": "II_inflow",
        "exInflow": "external_inflow",
        "flooding": "flooding",
        "outflow": "outflow",
        "evapLoss": "evaporation_loss",
        "seepLoss": "seepage_loss",
        "reacted": "reacted",
        "initStorage": "initial_storage",
        "finalStorage": "final_storage",
        "pctError": "routing_error",
    }


class RunoffTotals(ctypes.Structure):
    _fields_ = [
        ("rainfall", ctypes.c_double),
        ("evap", ctypes.c_double),
        ("infil", ctypes.c_double),
        ("runoff", ctypes.c_double),
        ("drains", ctypes.c_double),
        ("runon", ctypes.c_double),
        ("initStorage", ctypes.c_double),
        ("finalStorage", ctypes.c_double),
        ("initSnowCover", ctypes.c_double),
        ("finalSnowCover", ctypes.c_double),
        ("snowRemoved", ctypes.c_double),
        ("pctError", ctypes.c_double),
    ]
    _py_alias_ids = {
        "rainfall": "rainfall",
        "evap": "evaporation",
        "infil": "infiltration",
        "runoff": "runoff",
        "drains": "drains",
        "runon": "runon",
        "initStorage": "init_storage",
        "finalStorage": "final_storage",
        "initSnowCover": "init_snow_cover",
        "finalSnowCover": "final_snow_cover",
        "snowRemoved": "snow_removed",
        "pctError": "routing_error",
    }


# --- SWMM Output API
# -----------------------------------------------------------------------------
DLLErrorKeys = {
    411: "Input Error 411: no memory allocated for results.",
    412: "Input Error 412: no results; binary file hasn't been opened.",
    421: "Input Error 421: invalid parameter code.",
    434: "File Error  434: unable to open binary output file.",
    435: "File Error  435: run terminated; no results in binary file.",
    441: "Error 441: need to call SMR_open before calling this function",
}


class SMO_elementCount(Enum):
    subcatchCount = 0
    nodeCount = 1
    linkCount = 2
    pollutantCount = 3


class SMO_unit(Enum):
    flow_rate = 0
    concentration = 1


class SMO_apiFunction(Enum):
    getAttribute = 0
    getResult = 1


class SMO_elementType(Enum):
    SM_subcatch = 0
    SM_node = 1
    SM_link = 2
    SM_sys = 3


class SMO_time(Enum):
    reportStep = 0
    numPeriods = 1


class SMO_subcatchAttribute(Enum):
    rainfall_subcatch = 0
    snow_depth_subcatch = 1
    evap_loss = 2
    infil_loss = 3
    runoff_rate = 4
    gwoutflow_rate = 5
    gwtable_elev = 6
    soil_moisture = 7
    pollutant_conc_subcatch = 8


class SMO_nodeAttribute(Enum):
    invert_depth = 0
    hydraulic_head = 1
    stored_ponded_volume = 2
    lateral_inflow = 3
    total_inflow = 4
    flooding_losses = 5
    pollutant_conc_node = 6


class SMO_linkAttribute(Enum):
    flow_rate_link = 0
    flow_depth = 1
    flow_velocity = 2
    flow_volume = 3
    capacity = 4
    pollutant_conc_link = 5


class SMO_systemAttribute(Enum):
    air_temp = 0
    rainfall_system = 1
    snow_depth_system = 2
    evap_infil_loss = 3
    runoff_flow = 4
    dry_weather_inflow = 5
    groundwater_inflow = 6
    RDII_inflow = 7
    direct_inflow = 8
    total_lateral_inflow = 9
    flood_losses = 10
    outfall_flows = 11
    volume_stored = 12
    evap_rate = 13
